/* viewprog.c */

/* Based on: */
/***********************/
/* viewlace.c 1.00     */
/* for TestLace        */
/* by Adam M. Costello */
/* (amc@cs.wustl.edu)  */
/***********************/

/* This is ANSI C plus X11R5 code. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "jpeglib.h"


void fail(const char *msg)
{
  fputs(msg,stderr);
  exit(EXIT_FAILURE);
}


void warn(const char *msg)
{
  fputs(msg,stderr);
  fflush(stderr);
}

const char * const outofmem = "Out of memory.";



/******************* X11 interface *******************/

enum {depth = 8};

static const char *const
  cantopenwin = "Can't open window.\n";  /* Error message. */

static int width, height;

static char windowexists, gcexists, colormapexists;
  /* The above are initialized to 0. */
Display *display;
static Window window;
static Visual *visual;
static GC gc;
static Colormap colormap;


static void cleanup(void)
{
  if (gcexists) XFreeGC(display,gc);
  if (windowexists) XDestroyWindow(display,window);
  if (colormapexists) XFreeColormap(display,colormap);
}


void initwindow(const char *name, int w, int h)
{
  int screen;
  Window root;
  XVisualInfo visual_info;
  XSetWindowAttributes attributes;
  unsigned long valuemask;
  XSizeHints sizehints;
  XWMHints wmhints;
  XGCValues gcvalues;
  XEvent event, *ep = &event;

  atexit(cleanup);

  width  = w;
  height = h;

  screen = DefaultScreen(display);
  root   = RootWindow(display,screen);

  if (!XMatchVisualInfo(display, screen, depth, PseudoColor, &visual_info))
    fail("Screen does not support PseudoColor visual of desired depth.\n");
  visual = visual_info.visual;

  colormap = XCreateColormap(display,root,visual,AllocAll);
  if (!colormap) fail("Cannot create colormap.\n");
  colormapexists = 1;

  valuemask = CWBackingStore | CWEventMask;

  attributes.backing_store    = Always;
  attributes.event_mask       = ExposureMask | ButtonPressMask | KeyPressMask;

  window = XCreateWindow(display, root, 0, 0, width, height, 0, depth,
                         InputOutput, visual, valuemask, &attributes);
  if (window == None) fail(cantopenwin);
  windowexists = 1;

  XSetWindowColormap(display,window,colormap);

  sizehints.flags = PMinSize | PMaxSize | PBaseSize;
  sizehints.min_width  = sizehints.max_width  = sizehints.base_width  = width;
  sizehints.min_height = sizehints.max_height = sizehints.base_height = height;
  XSetWMNormalHints(display, window, &sizehints);

  wmhints.flags = InputHint | StateHint;
  wmhints.input = True;
  wmhints.initial_state = NormalState;
  XSetWMHints(display, window, &wmhints);

  XStoreName(display,window,name);

  XMapWindow(display,window);

  valuemask = 0;
  gc = XCreateGC(display, window, valuemask, &gcvalues);
  gcexists = 1;

  do XNextEvent(display,ep);
  while (   ep->xany.window != window
         || ep->type != Expose || ep->xexpose.count);

  XFlush(display);
}


void setName (const char *name, int pass)
{
  char tname[1000];

  sprintf(tname, "%s (%d)", name, pass);

  XStoreName(display,window,tname);
}


static unsigned long pixels[1 << depth];


void definecolor (int code, unsigned short red,
		  unsigned short green, unsigned short blue)
{
  XColor xcolor;
  static int firstcall = 1;

  xcolor.pixel = code;
  xcolor.red   = red;
  xcolor.green = green;
  xcolor.blue  = blue;
  xcolor.flags = DoRed | DoGreen | DoBlue;
  if (!XStoreColor(display, colormap, &xcolor))
    fail("Unable to set colormap entry.\n");
  pixels[code] = xcolor.pixel;
  if (firstcall) XFillRectangle(display,window,gc,0,0,width,height);
  firstcall = 0;
}


void drawscanline(int y, unsigned char *scanline)
{
  static XImage * xim = NULL;
  static unsigned char *imagedata;
  int x;

  if (!xim) {
    /* This code borrowed from XV */
    int   imWIDTH, nullCount;
  
    nullCount = (4 - (width % 4)) & 0x03;  /* # of padding bytes per line */
    imWIDTH = width + nullCount;
 
    /* Now create the image data - pad each scanline as necessary */
    imagedata = (unsigned char *) malloc((size_t) imWIDTH);
    if (!imagedata) fail("couldn't malloc imagedata");
    memset(imagedata, 0, imWIDTH);
      
    xim = XCreateImage(display,visual,depth,ZPixmap,0,
		       (char *) imagedata,  width,  1,
		       32, imWIDTH);
    if (!xim) fail("XCreateImage failed");
  }

  for (x = 0; x < width; x++) {
    imagedata[x] = (unsigned char) pixels[scanline[x]];
  }

  XPutImage(display, window, gc, xim, 0,0, 0,y, width,1);
}


/******************* end X11 interface *******************/



void define_colormap (j_decompress_ptr cinfo)
{
  int code;
  unsigned short R, G, B;
  JSAMPARRAY colormap = cinfo->colormap;

  if (cinfo->quantize_colors) {
    for (code = 0; code < cinfo->actual_number_of_colors; code++) {
      R = colormap[0][code];
      G = colormap[1][code];
      B = colormap[2][code];
      definecolor(code, (R << 8) | R, (G << 8) | G, (B << 8) | B);
    }
  } else {
    /* Grayscale */
    for (code = 0; code < 256; code++) {
      R = G = B = code;
      definecolor(code, (R << 8) | R, (G << 8) | G, (B << 8) | B);
    }
  }
}


void eventloop(void)
/* Just waits for a keystroke or button 1 press, then returns. */
{
  XEvent event;

  for (;;) {
    XNextEvent(display,&event);
    if (event.type == ButtonPress && event.xbutton.button == Button1)
      break;
    if (event.type == KeyPress)
      break;
  }
}


static int
keymatch (const char * arg, const char * keyword, int minchars)
/* Case-insensitive matching of (possibly abbreviated) keyword switches. */
/* keyword is the constant keyword (must be lower case already), */
/* minchars is length of minimum legal abbreviation. */
{
  register int ca, ck;
  register int nmatched = 0;

  while ((ca = *arg++) != '\0') {
    if ((ck = *keyword++) == '\0')
      return 0;			/* arg longer than keyword, no good */
    if (isupper(ca))		/* force arg to lcase (assume ck is already) */
      ca = tolower(ca);
    if (ca != ck)
      return 0;			/* no good */
    nmatched++;			/* count matched characters */
  }
  /* reached end of argument; fail if it's too short for unique abbrev */
  if (nmatched < minchars)
    return 0;
  return 1;			/* A-OK */
}


void usage(void)
/* Prints a usage message to stderr, then exits unsuccessfully. */
{
  fputs(
    "Options:\n"
    "  -display <displayname>   Choose X display\n"
    "  -name    <winname>       Set window name (default is input file name)\n"
    "  -colors  <n>             Number of colors to use (default 256)\n"
    "  -fast                    Use fast/sloppy processing, except for final pass\n"
    "  -smooth                  Use block smoothing\n"
    "  -testskip                Skip alternate scans (for testing only)\n"
    "  -verbose                 Verbose library output (more -v's = even more)\n"
    "  -wait                    Wait for click/keystroke after each scan\n"
    "  -2                       Use 2-pass quantization for final pass\n"
    "  <filename>               Read image from a file (default is stdin)\n"
    , stderr);

  fail(
    "\n"
    );
}


int main(int argc, const char **argv)
{
  const char **argp = argv;
  const char *displayname = NULL, *winname = NULL, *infilename = NULL;
  boolean usefast = FALSE;
  boolean use2pass = FALSE;
  boolean scanwait = FALSE;
  boolean dosmooth = FALSE;
  boolean skipalt = FALSE;
  int numcolors = 256;
  int verbosity = 0;
  FILE *infile = NULL;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  int width, height;
  JSAMPLE *buffer[1];
  boolean skipped_last = TRUE;
  boolean final_pass;

  while (*++argp) {
    if        (keymatch(*argp, "-colors", 2)) {
      if (!*++argp) usage();
      numcolors = atoi(*argp);
    } else if (keymatch(*argp, "-display", 2)) {
      if (!*++argp) usage();
      displayname = *argp;
    } else if (keymatch(*argp, "-fast", 2)) {
      usefast = TRUE;
    } else if (keymatch(*argp, "-name", 2)) {
      if (!*++argp) usage();
      winname = *argp;
    } else if (keymatch(*argp, "-smooth", 2)) {
      dosmooth = TRUE;
    } else if (keymatch(*argp, "-testskip", 2)) {
      skipalt = TRUE;
    } else if (keymatch(*argp, "-verbose", 2)) {
      verbosity++;
    } else if (keymatch(*argp, "-wait", 2)) {
      scanwait = TRUE;
    } else if (keymatch(*argp, "-2", 2)) {
      use2pass = TRUE;
    } else if (**argp != '-') {
      infilename = *argp;
      if (argp[1]) usage();
    } else {
      usage();
    }
  }

  if (!infilename) {
    infile = stdin;
    if (!winname) winname = argv[0];
  } else {
    infile = fopen(infilename, "rb");
    if (!infile) fail("Cannot open input file.\n");
    if (!winname) winname = infilename;
  }

  /* Initialize the JPEG decompression object with default error handling. */
  cinfo.err = jpeg_std_error(&jerr);
  jerr.trace_level = verbosity;
  jpeg_create_decompress(&cinfo);

  /* Specify data source for decompression */
  jpeg_stdio_src(&cinfo, infile);

  /* Read file header, set default decompression parameters */
  (void) jpeg_read_header(&cinfo, TRUE);

  width = cinfo.image_width;
  height = cinfo.image_height;

  buffer[0] = malloc(width * sizeof(JSAMPLE));
  if (!buffer[0]) fail(outofmem);

  display = XOpenDisplay(displayname);
  if (!display) fail("Cannot open display.\n");

  initwindow(winname,width,height);

  /* Start decompressor */
  if (cinfo.num_components > 1)
    cinfo.quantize_colors = TRUE;
  cinfo.two_pass_quantize = FALSE;
  cinfo.desired_number_of_colors = numcolors;
  if (usefast) {
    cinfo.dct_method = JDCT_FASTEST;
    cinfo.dither_mode = JDITHER_ORDERED;
    cinfo.do_fancy_upsampling = FALSE;
  }
  if (use2pass)
    cinfo.enable_2pass_quant = TRUE;
  cinfo.do_block_smoothing = dosmooth;
  cinfo.buffered_image = TRUE;
  (void) jpeg_start_decompress(&cinfo);

  do {
    final_pass = jpeg_input_complete(&cinfo);
    /* Skip alternate scans if enabled; this is mostly for test purposes */
    if (skipalt && !final_pass && (cinfo.input_scan_number & 1) == 0) {
      if (verbosity>0)
	fprintf(stderr, "Skipping scan %d row %u\n",
		cinfo.input_scan_number, cinfo.input_iMCU_row);
      for (;;) {
	int retcode = jpeg_consume_input(&cinfo);
	if (retcode == JPEG_SUSPENDED)
	  fprintf(stderr, "jpeg_consume_input suspended!\n");
	else if (retcode != JPEG_ROW_COMPLETED)
	  break;
      }
      if (verbosity>1)
	fprintf(stderr, "Done at scan %d row %u\n",
		cinfo.input_scan_number, cinfo.input_iMCU_row);
      skipped_last = TRUE;
    } else {
      if (final_pass) {
	if (!skipped_last && !use2pass)
	  break;		/* don't make an extra display pass */
	/* increase quality for final pass */
	cinfo.dct_method = JDCT_ISLOW;
	cinfo.dither_mode = JDITHER_FS;
	cinfo.two_pass_quantize = use2pass;
	if (use2pass) cinfo.colormap = NULL;
      }
      if (! jpeg_start_output(&cinfo, cinfo.input_scan_number))
	fprintf(stderr, "jpeg_start_output returned FALSE!\n");
      setName(winname, cinfo.input_scan_number);
      define_colormap(&cinfo);

      /* Process data */
      while (cinfo.output_scanline < cinfo.output_height) {
	if (! jpeg_read_scanlines(&cinfo, buffer, 1))
	  fprintf(stderr, "jpeg_read_scanlines returned 0!\n");
	drawscanline(cinfo.output_scanline-1, buffer[0]);
      }

      if (! jpeg_finish_output(&cinfo))
	fprintf(stderr, "jpeg_finish_output returned FALSE!\n");
      skipped_last = FALSE;
      XFlush(display);

      if (scanwait) {
	putchar('\007');
	fflush(stdout);
	eventloop();
      }
    }
  } while (! final_pass);

  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  if (infile != stdin)
    fclose(infile);

  XFlush(display);
  puts("\007Done.  Click or type in window to exit.");
  fflush(stdout);

  eventloop();

  return 0;
}
