/* viewrate.c */

/* Based on: */
/***********************/
/* viewlace.c 1.00     */
/* for TestLace        */
/* by Adam M. Costello */
/* (amc@cs.wustl.edu)  */
/***********************/

/* This is ANSI C plus X11R5 code. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "jpeglib.h"


void fail(const char *msg)
{
  fputs(msg,stderr);
  exit(EXIT_FAILURE);
}


void warn(const char *msg)
{
  fputs(msg,stderr);
  fflush(stderr);
}

const char * const outofmem = "Out of memory.";


/******************* Simulation of constant-data-rate input ***************/

/* Expanded data source object for stdio input */

typedef struct {
  struct jpeg_source_mgr pub;	/* public fields */

  FILE * infile;		/* source stream */
  JOCTET * buffer;		/* start of buffer */
  size_t bufsize;		/* physical buffer size */
  size_t chunksize;		/* number of bytes to read at once */
  int inputrate;		/* allowed bytes/sec */
  struct timeval starttime;	/* time at start of run */
  long total_read;		/* total bytes read so far */
  long remaining_skip;		/* bytes remaining to skip */
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;

#define MIN_BUF_SIZE  4096	/* assume library leaves no more than this in buf */


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

METHODDEF void
init_source (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;
  struct timezone tzp;

  gettimeofday(&src->starttime, &tzp);
  src->total_read = 0;
  src->remaining_skip = 0;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 */

METHODDEF boolean
fill_input_buffer (j_decompress_ptr cinfo)
{
  return FALSE;			/* force suspension */
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 */

METHODDEF void
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  if (num_bytes > 0) {
    if (num_bytes > (long) src->pub.bytes_in_buffer) {
      src->remaining_skip += num_bytes - (long) src->pub.bytes_in_buffer;
      src->pub.bytes_in_buffer = 0;
    } else {
      src->pub.next_input_byte += (size_t) num_bytes;
      src->pub.bytes_in_buffer -= (size_t) num_bytes;
    }
  }
}


/*
 * Read more data --- called frequently, reads data at proper rate.
 */

void
receive_more_data (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;
  size_t nbytes, alreadyinbuf;
  struct timeval now;		/* current time */
  struct timezone tzp;
  long elapsed, targtime;

  /* Compute elapsed time since start of run, in microsec */
  gettimeofday(&now, &tzp);
  if (src->starttime.tv_usec > now.tv_usec) {
    now.tv_usec += 1000000L;
    now.tv_sec--;
  }
  elapsed = (now.tv_sec - src->starttime.tv_sec) * 1000000L;
  elapsed += now.tv_usec - src->starttime.tv_usec;

  /* Time to read another chunk? */
  targtime = (long) ((src->total_read + src->chunksize) * 1000000.0 /
		     src->inputrate);
  if (targtime > elapsed) return;

  /* Move any existing data down to front of buffer */
  alreadyinbuf = src->pub.bytes_in_buffer;
  if (alreadyinbuf > 0)
    memmove(src->buffer, src->pub.next_input_byte,
	    alreadyinbuf * sizeof(JOCTET));
  src->pub.next_input_byte = src->buffer;

  /* Read up to chunksize bytes */
  nbytes = src->bufsize - alreadyinbuf; /* max space avail */
  if (nbytes > src->chunksize)
    nbytes = src->chunksize;

  nbytes = fread(src->buffer + alreadyinbuf,
		 (size_t) 1, nbytes, src->infile);
  src->pub.bytes_in_buffer += nbytes;
  src->total_read += nbytes;

  if (src->remaining_skip > 0) {
    if (src->remaining_skip > (long) src->pub.bytes_in_buffer) {
      src->remaining_skip -= (long) src->pub.bytes_in_buffer;
      src->pub.bytes_in_buffer = 0;
    } else {
      src->pub.next_input_byte += (size_t) src->remaining_skip;
      src->pub.bytes_in_buffer -= (size_t) src->remaining_skip;
      src->remaining_skip = 0;
    }
  }
}



/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 */

METHODDEF void
term_source (j_decompress_ptr cinfo)
{
  /* no work necessary here */
}


/*
 * Prepare for input from a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

void
rate_stdio_src (j_decompress_ptr cinfo, FILE * infile,
		int inputrate, int chunksize)
{
  my_src_ptr src;
  size_t bufsize;

  if (inputrate < 1) inputrate = 1;
  if (chunksize < 100) chunksize = 100;
  bufsize = chunksize + MIN_BUF_SIZE;

  if (cinfo->src == NULL) {	/* first time for this JPEG object? */
    cinfo->src = (struct jpeg_source_mgr *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  sizeof(my_source_mgr));
    src = (my_src_ptr) cinfo->src;
    src->buffer = (JOCTET *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  bufsize * sizeof(JOCTET));
    src->bufsize = bufsize;
    src->chunksize = chunksize;
    src->inputrate = inputrate;
  }

  src = (my_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
  src->pub.term_source = term_source;
  src->infile = infile;
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
  src->pub.next_input_byte = src->buffer;
}


/******************* X11 interface *******************/

enum {depth = 8};

static const char *const
  cantopenwin = "Can't open window.\n";  /* Error message. */

static int width, height;

static char windowexists, gcexists, colormapexists;
  /* The above are initialized to 0. */
Display *display;
static Window window;
static Visual *visual;
static GC gc;
static Colormap colormap;


static void cleanup(void)
{
  if (gcexists) XFreeGC(display,gc);
  if (windowexists) XDestroyWindow(display,window);
  if (colormapexists) XFreeColormap(display,colormap);
}


void initwindow(const char *name, int w, int h)
{
  int screen;
  Window root;
  XVisualInfo visual_info;
  XSetWindowAttributes attributes;
  unsigned long valuemask;
  XSizeHints sizehints;
  XWMHints wmhints;
  XGCValues gcvalues;
  XEvent event, *ep = &event;

  atexit(cleanup);

  width  = w;
  height = h;

  screen = DefaultScreen(display);
  root   = RootWindow(display,screen);

  if (!XMatchVisualInfo(display, screen, depth, PseudoColor, &visual_info))
    fail("Screen does not support PseudoColor visual of desired depth.\n");
  visual = visual_info.visual;

  colormap = XCreateColormap(display,root,visual,AllocAll);
  if (!colormap) fail("Cannot create colormap.\n");
  colormapexists = 1;

  valuemask = CWBackingStore | CWEventMask;

  attributes.backing_store    = Always;
  attributes.event_mask       = ExposureMask | ButtonPressMask | KeyPressMask;

  window = XCreateWindow(display, root, 0, 0, width, height, 0, depth,
                         InputOutput, visual, valuemask, &attributes);
  if (window == None) fail(cantopenwin);
  windowexists = 1;

  XSetWindowColormap(display,window,colormap);

  sizehints.flags = PMinSize | PMaxSize | PBaseSize;
  sizehints.min_width  = sizehints.max_width  = sizehints.base_width  = width;
  sizehints.min_height = sizehints.max_height = sizehints.base_height = height;
  XSetWMNormalHints(display, window, &sizehints);

  wmhints.flags = InputHint | StateHint;
  wmhints.input = True;
  wmhints.initial_state = NormalState;
  XSetWMHints(display, window, &wmhints);

  XStoreName(display,window,name);

  XMapWindow(display,window);

  valuemask = 0;
  gc = XCreateGC(display, window, valuemask, &gcvalues);
  gcexists = 1;

  do XNextEvent(display,ep);
  while (   ep->xany.window != window
         || ep->type != Expose || ep->xexpose.count);

  XFlush(display);
}


void setName (const char *name, int pass)
{
  char tname[1000];

  sprintf(tname, "%s (%d)", name, pass);

  XStoreName(display,window,tname);
}


static unsigned long pixels[1 << depth];


void definecolor (int code, unsigned short red,
		  unsigned short green, unsigned short blue)
{
  XColor xcolor;
  static int firstcall = 1;

  xcolor.pixel = code;
  xcolor.red   = red;
  xcolor.green = green;
  xcolor.blue  = blue;
  xcolor.flags = DoRed | DoGreen | DoBlue;
  if (!XStoreColor(display, colormap, &xcolor))
    fail("Unable to set colormap entry.\n");
  pixels[code] = xcolor.pixel;
  if (firstcall) XFillRectangle(display,window,gc,0,0,width,height);
  firstcall = 0;
}


void drawscanline(int y, unsigned char *scanline)
{
  static XImage * xim = NULL;
  static unsigned char *imagedata;
  int x;

  if (!xim) {
    /* This code borrowed from XV */
    int   imWIDTH, nullCount;
  
    nullCount = (4 - (width % 4)) & 0x03;  /* # of padding bytes per line */
    imWIDTH = width + nullCount;
 
    /* Now create the image data - pad each scanline as necessary */
    imagedata = (unsigned char *) malloc((size_t) imWIDTH);
    if (!imagedata) fail("couldn't malloc imagedata");
    memset(imagedata, 0, imWIDTH);
      
    xim = XCreateImage(display,visual,depth,ZPixmap,0,
		       (char *) imagedata,  width,  1,
		       32, imWIDTH);
    if (!xim) fail("XCreateImage failed");
  }

  for (x = 0; x < width; x++) {
    imagedata[x] = (unsigned char) pixels[scanline[x]];
  }

  XPutImage(display, window, gc, xim, 0,0, 0,y, width,1);
}


/******************* end X11 interface *******************/



void define_colormap (j_decompress_ptr cinfo)
{
  int code;
  unsigned short R, G, B;
  JSAMPARRAY colormap = cinfo->colormap;

  if (cinfo->quantize_colors) {
    for (code = 0; code < cinfo->actual_number_of_colors; code++) {
      R = colormap[0][code];
      G = colormap[1][code];
      B = colormap[2][code];
      definecolor(code, (R << 8) | R, (G << 8) | G, (B << 8) | B);
    }
  } else {
    /* Grayscale */
    for (code = 0; code < 256; code++) {
      R = G = B = code;
      definecolor(code, (R << 8) | R, (G << 8) | G, (B << 8) | B);
    }
  }
}


void eventloop(void)
/* Just waits for a keystroke or button 1 press, then returns. */
{
  XEvent event;

  for (;;) {
    XNextEvent(display,&event);
    if (event.type == ButtonPress && event.xbutton.button == Button1)
      break;
    if (event.type == KeyPress)
      break;
  }
}


static int
keymatch (const char * arg, const char * keyword, int minchars)
/* Case-insensitive matching of (possibly abbreviated) keyword switches. */
/* keyword is the constant keyword (must be lower case already), */
/* minchars is length of minimum legal abbreviation. */
{
  register int ca, ck;
  register int nmatched = 0;

  while ((ca = *arg++) != '\0') {
    if ((ck = *keyword++) == '\0')
      return 0;			/* arg longer than keyword, no good */
    if (isupper(ca))		/* force arg to lcase (assume ck is already) */
      ca = tolower(ca);
    if (ca != ck)
      return 0;			/* no good */
    nmatched++;			/* count matched characters */
  }
  /* reached end of argument; fail if it's too short for unique abbrev */
  if (nmatched < minchars)
    return 0;
  return 1;			/* A-OK */
}


void usage(void)
/* Prints a usage message to stderr, then exits unsuccessfully. */
{
  fputs(
    "Options:\n"
    "  -display <displayname>   Choose X display\n"
    "  -name    <winname>       Set window name (default is input file name)\n"
    "  -rate    <n>             Input data rate in bytes/sec (default 1000)\n"
    "  -chunk   <n>             Input chunk size in bytes (default 1024)\n"
    "  -colors  <n>             Number of colors to use (default 256)\n"
    "  -fast                    Use fast/sloppy processing, except for final pass\n"
    "  -smooth                  Use block smoothing\n"
    "  -verbose                 Verbose library output (more -v's = even more)\n"
    "  -2                       Use 2-pass quantization for final pass\n"
    "  <filename>               Read image from a file (default is stdin)\n"
    , stderr);

  fail(
    "\n"
    );
}


int main(int argc, const char **argv)
{
  const char **argp = argv;
  const char *displayname = NULL, *winname = NULL, *infilename = NULL;
  boolean usefast = FALSE;
  boolean use2pass = FALSE;
  boolean dosmooth = FALSE;
  int inputrate = 1000;
  int chunksize = 1024;
  int numcolors = 256;
  int verbosity = 0;
  FILE *infile = NULL;
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  int width, height;
  int retval;
  JSAMPLE *buffer[1];
  boolean final_pass;

  while (*++argp) {
    if        (keymatch(*argp, "-colors", 3)) {
      if (!*++argp) usage();
      numcolors = atoi(*argp);
    } else if (keymatch(*argp, "-chunk", 3)) {
      if (!*++argp) usage();
      chunksize = atoi(*argp);
    } else if (keymatch(*argp, "-display", 2)) {
      if (!*++argp) usage();
      displayname = *argp;
    } else if (keymatch(*argp, "-fast", 2)) {
      usefast = TRUE;
    } else if (keymatch(*argp, "-name", 2)) {
      if (!*++argp) usage();
      winname = *argp;
    } else if (keymatch(*argp, "-rate", 2)) {
      if (!*++argp) usage();
      inputrate = atoi(*argp);
    } else if (keymatch(*argp, "-smooth", 2)) {
      dosmooth = TRUE;
    } else if (keymatch(*argp, "-verbose", 2)) {
      verbosity++;
    } else if (keymatch(*argp, "-2", 2)) {
      use2pass = TRUE;
    } else if (**argp != '-') {
      infilename = *argp;
      if (argp[1]) usage();
    } else {
      usage();
    }
  }

  if (!infilename) {
    infile = stdin;
    if (!winname) winname = argv[0];
  } else {
    infile = fopen(infilename, "rb");
    if (!infile) fail("Cannot open input file.\n");
    if (!winname) winname = infilename;
  }

  /* Initialize the JPEG decompression object with default error handling. */
  cinfo.err = jpeg_std_error(&jerr);
  jerr.trace_level = verbosity;
  jpeg_create_decompress(&cinfo);

  /* Specify data source for decompression */
  rate_stdio_src(&cinfo, infile, inputrate, chunksize);

  /* Read file header, set default decompression parameters */
  while (jpeg_read_header(&cinfo, TRUE) == JPEG_SUSPENDED)
    receive_more_data(&cinfo);

  width = cinfo.image_width;
  height = cinfo.image_height;

  buffer[0] = malloc(width * sizeof(JSAMPLE));
  if (!buffer[0]) fail(outofmem);

  display = XOpenDisplay(displayname);
  if (!display) fail("Cannot open display.\n");

  initwindow(winname,width,height);

  /* Start decompressor */
  if (cinfo.num_components > 1)
    cinfo.quantize_colors = TRUE;
  cinfo.two_pass_quantize = FALSE;
  cinfo.desired_number_of_colors = numcolors;
  if (usefast) {
    cinfo.dct_method = JDCT_FASTEST;
    cinfo.dither_mode = JDITHER_ORDERED;
    cinfo.do_fancy_upsampling = FALSE;
  }
  if (use2pass)
    cinfo.enable_2pass_quant = TRUE;
  cinfo.do_block_smoothing = dosmooth;
  cinfo.buffered_image = TRUE;

  if (! jpeg_start_decompress(&cinfo))
    fail("unexpected suspension return from start_decompress()\n");

  do {
    /* Absorb any pending data before checking for EOF */
    do {
      receive_more_data(&cinfo);
      retval = jpeg_consume_input(&cinfo);
    } while (retval != JPEG_SUSPENDED && retval != JPEG_REACHED_EOI);
    final_pass = jpeg_input_complete(&cinfo);
    if (final_pass) {
      /* increase quality for final pass */
      cinfo.dct_method = JDCT_ISLOW;
      cinfo.dither_mode = JDITHER_FS;
      cinfo.two_pass_quantize = use2pass;
      if (use2pass) cinfo.colormap = NULL;
    }
    while (! jpeg_start_output(&cinfo, cinfo.input_scan_number))
      receive_more_data(&cinfo);
    setName(winname, cinfo.input_scan_number);
    define_colormap(&cinfo);

    /* Process data */
    while (cinfo.output_scanline < cinfo.output_height) {
      /* Absorb any pending data */
      do {
	receive_more_data(&cinfo);
	retval = jpeg_consume_input(&cinfo);
      } while (retval != JPEG_SUSPENDED && retval != JPEG_REACHED_EOI);
      /* Abort nonfinal pass if input EOF reached */
      if (retval == JPEG_REACHED_EOI && !final_pass)
	break;
      /* Emit output if available */
      if (jpeg_read_scanlines(&cinfo, buffer, 1))
	drawscanline(cinfo.output_scanline-1, buffer[0]);
    }

    while (! jpeg_finish_output(&cinfo))
      receive_more_data(&cinfo);
  } while (! final_pass);

  if (! jpeg_finish_decompress(&cinfo))
    warn("unexpected suspension return from finish_decompress()\n");
  jpeg_destroy_decompress(&cinfo);

  if (infile != stdin)
    fclose(infile);

  XFlush(display);
  puts("\007Done.  Click or type in window to exit.");
  fflush(stdout);

  eventloop();

  return 0;
}
